/* AbulEduCameraControllerV1
 *
 * @warning aucun traitement d'erreur n'est pour l'instant implémenté
 * @see https://redmine.ryxeo.com/projects/leterrier-developpeu/repository
 * @author 2015 Jean-Louis Frucot <frucot.jeanlouis@free.fr>
 * @see The GNU Public License (GPL)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "abuleducameracontrollerv1.h"

AbulEduCameraControllerV1::AbulEduCameraControllerV1(QObject *parent) :
    QObject(parent),
    m_camWorker(0),
    m_camThread(0),
    m_camTimer(0),
    m_globalTime(0),
    m_deviceNumber(-2),
    m_resWidth(DEFAULT_HORIZONTAL_RES),
    m_resHeight(DEFAULT_VERTICAL_RES),
    m_frameRefreshInterval(DEFAULT_REFRESH_INTERVAL),
    m_flagReady(ImageSend)

{
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__<< "---------- Constructeur ----------";

    m_localDebug = true;
    m_camTimer   = new QTimer(this);
    m_globalTime = new QTime();
    sampleNumber = 0;
    fpsSum       = 0;
    m_fpsQueue.clear();

    createCameraThread();
}

AbulEduCameraControllerV1::~AbulEduCameraControllerV1()
{
    if(m_camWorker)
    {
        m_camWorker->deleteLater();
    }
    /* On détruit proprement le thread */
    if(m_camThread)
    {
        m_camThread->quit();
        m_camThread->wait();
    }
    delete m_globalTime;
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__<< "---------- Destructeur ----------";
}

void AbulEduCameraControllerV1::abeCameraInitCamera(int deviceNumber, int resWidth, int resHeight)
{
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__;

    if(m_deviceNumber != deviceNumber) // On change de caméra
    {
        if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__<<"On change de caméra";
        m_deviceNumber = deviceNumber;
        m_resWidth     = resWidth;
        m_resHeight    = resHeight;
        abeCameraSetFrameResolution(resWidth, resHeight);

        emit signalSetDevice(m_deviceNumber, m_resWidth, m_resHeight);
    }
    else // La résolution a peut-être changé
    {
        if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__<<"On ne change pas de caméra";
        abeCameraSetFrameResolution(resWidth, resHeight);
    }
}

void AbulEduCameraControllerV1::abeCameraSetFrameResolution(int width, int height)
{
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__<<width<<"x"<<height;
    if(m_resHeight != height || m_resWidth != width) // La résolution a changé
    {
        m_resHeight = height;
        m_resWidth = width;
    }
    emit signalSetImageResolution(m_resWidth, m_resHeight);
}

void AbulEduCameraControllerV1::abeCameraSetFrameRefreshInterval(int refreshInterval)
{
//    m_camTimer->stop(); // pas nécessaire
    m_camTimer->setInterval(refreshInterval);
}

void AbulEduCameraControllerV1::abeCameraSetROI(int x, int y, int width, int height)
{
//    emit signal
}

void AbulEduCameraControllerV1::abeCameraStartCapture(int delai)
{
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__;
    Q_ASSERT_X(delai >=0, __FUNCTION__, "le delai doit être supérieur ou égal à 0");
    m_camTimer->stop();
    m_flagReady = Ready;
    QTimer::singleShot(delai, m_camTimer, SLOT(start()));
}

void AbulEduCameraControllerV1::abeCameraStopCapture(int delai)
{
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__;
    Q_ASSERT_X(delai >=0, __FUNCTION__, "le delai doit être supérieur ou égal à 0");
    QTimer::singleShot(delai, m_camTimer, SLOT(stop()));
    m_flagReady &= ~ImageNeeded; // On met le flag à 0
}

void AbulEduCameraControllerV1::createCameraThread()
{
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__;
    m_camThread = new QThread(this);
    m_camWorker = new AbulEduCameraWorker();
    m_camTimer = new QTimer(this);
    /* On établit la communication avec le camWorker */
    connect(this,           SIGNAL(signalSetDevice(int, int, int)),
            m_camWorker, SLOT(slotAbeCamWorkerSetDevice(int, int, int)));

    connect(this,           SIGNAL(signalSetImageResolution(int, int)),
            m_camWorker,    SLOT(slotAbeCamWorkerSetResolution(int, int)));

    connect(m_camTimer,     SIGNAL(timeout()),
            this,           SLOT(abeCameraGetFullImage()));

    connect(m_camWorker,    SIGNAL(signalAbeCamWorkerQImageReady(QImage)),
            this,           SLOT(slotQImageFromCamReceived(QImage)));

    connect(m_camWorker,    SIGNAL(signalAbeCamWorkerImageSize(int, int)),
            this,           SIGNAL(signalAbeCameraImageSize(int,int)));

    connect(m_camWorker,    SIGNAL(signalAbeCamWorkerImageCvMatReady(cv::Mat)),
            this,           SIGNAL(signalFullCvMatImageReady(cv::Mat)));

    connect(m_camWorker,    SIGNAL(signalAbeCamWorkerImageSize(int, int)),
            this,           SLOT(slotGetImageSize(int, int)));

    connect(m_camWorker,    SIGNAL(signalAbeCamWorkerError(int)),
            this,           SLOT(slotCameraWorkerErrorHandler(int)));

    connect(m_camWorker,    SIGNAL(signalAbeCamWorkerError(int)),
            this,           SIGNAL(signalAbeCameraError(int)));

    connect(this,           SIGNAL(signalGetFullImage()),
            m_camWorker,    SLOT(slotAbeCamWorkerGetFullImage()));

    connect(this,           SIGNAL(signalGetFullImage()),
            m_camWorker,    SLOT(slotAbeCamWorkerGetFullCvMatImage()));

    /* On place la "camera" dans un thread indépendant */
    m_camWorker->moveToThread(m_camThread);
    /* Et on démarre le thread */
    m_camThread->start();
}

void AbulEduCameraControllerV1::slotQImageFromCamReceived(QImage image)
{
    //    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__;
    if(m_flagReady.testFlag(ImageNeeded))
    {
        /* On est pret à recevoir une autre image
         * Mais il faut attendre qu'elle soit demandée
         */
        m_flagReady = ImageSend;
        emit signalFullImageReady(image);
    }
}

void AbulEduCameraControllerV1::slotCvMatFromCamReceived(cv::Mat frame)
{
    if(m_flagReady.testFlag(ImageNeeded))
    {
        /* On est pret à recevoir une autre image
         * Mais il faut attendre qu'elle soit demandée
         */
        m_flagReady = ImageSend;
        emit signalFullCvMatImageReady(frame);
    }
}

void AbulEduCameraControllerV1::slotGetImageSize(int width, int height)
{
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__;
    // Pour l'instant, le signal est ré-émis directement par un connect

}

void AbulEduCameraControllerV1::abeCameraGetFullImage()
{
//    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__<<__LINE__;
    /* Si une image a été traitée, on peut passer à la suivante
     * Cela évite d'avoir des appels à abuleducameraworker qui s'empilent sans espoir !!!!!
     */
    if(m_flagReady.testFlag(Ready))
    {
        emit signalGetFullImage();
        m_flagReady &= ~ImageSend;
    }
}

void AbulEduCameraControllerV1::slotCameraWorkerErrorHandler(int errorNumber)
{
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__<< "Erreur N°"<<errorNumber;

    switch (errorNumber)
    {
    case 0: // Pas d'erreur, on peut démarrer le timer de la camera
        break;
    case 1:
        m_camTimer->stop();
        break;
    default:
        break;
    }
}

void AbulEduCameraControllerV1::slotCameraControllerReadyToGetNewImage(bool ready)
{
//    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__;
    m_elapsedGlobalTime = m_globalTime->elapsed();
    m_globalTime->start();
    updateFPS(m_elapsedGlobalTime);
    if(ready)
    {
        m_flagReady |= ImageNeeded;
    }
    else
    {
        m_flagReady &= ~ImageNeeded;
    }
}

void AbulEduCameraControllerV1::updateFPS(int timeElapsed)
{
//    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__;
    // On pousse le temps passé sur la pile
    if(timeElapsed>0)
    {
        m_fpsQueue.enqueue((int)1000/timeElapsed);
        // Incremente le nombre d'images
        sampleNumber++;
    }
    // La taille maximum de la pile est DEFAULT_CAPTURE_FPS_STAT_QUEUE_LENGTH
    if(m_fpsQueue.size()>CAPTURE_FPS_STAT_QUEUE_LENGTH)
        m_fpsQueue.dequeue();
    // On met à jour la valeur moyenne toutes les CAPTURE_FPS_STAT_QUEUE_LENGTH images
    if((m_fpsQueue.size()==CAPTURE_FPS_STAT_QUEUE_LENGTH)&&(sampleNumber==CAPTURE_FPS_STAT_QUEUE_LENGTH))
    {
        // On vide la pile en calculant la moyenne
        while(!m_fpsQueue.empty())
        {
            fpsSum+=m_fpsQueue.dequeue();
        }
        // Calcul de la moyenne -> FPS
        emit signalCameraFPSAverage(fpsSum/CAPTURE_FPS_STAT_QUEUE_LENGTH);
        // RAZ
        fpsSum=0;
        sampleNumber=0;
    }
}
