/* AbulEduCameraWorkerV1
 *
 * @warning aucun traitement d'erreur n'est pour l'instant implémenté
 * @see https://redmine.ryxeo.com/projects/leterrier-developpeu/repository
 * @author 2015 Jean-Louis Frucot <frucot.jeanlouis@free.fr>
 * @see The GNU Public License (GPL)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "abuleducameraworkerv1.h"

AbulEduCameraWorker::AbulEduCameraWorker(QObject *parent) :
    QObject(parent),
    m_localDebug(true),
    m_isCameraConnected(false),
    m_deviceNumber(-2)
{
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__<< "---------- Constructeur ----------";
    /* Normalement,on ne fait pas d'initialisations ici car cette classe est destinée à tourner dans un thread séparé
     * La communication avec la classe se fait uniquement par le biais de signaux et de slots qui sont thread-safe
     * Le point d'entrée est slotAbeCamWorkerSetDevice(int devicenumber)
     */
}

AbulEduCameraWorker::~AbulEduCameraWorker()
{
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__<< "---------- Destructeur ----------";
    delete m_image;
}

void AbulEduCameraWorker::slotAbeCamWorkerSetDevice(int deviceNumber, int width, int height)
{
     qRegisterMetaType< cv::Mat >("cv::Mat"); // Permet de passer des cv::mat dans les signaux
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__<<__LINE__<<"Device N°"<<deviceNumber<<width<<height;

    if(m_deviceNumber == deviceNumber) // On n'a pas changé de caméra
    {
        if(m_resWidth != width || m_resHeight != height) // Mais on a changé de résolution (ou pas)
        {
            m_resWidth = width;
            m_resHeight = height;
        }
        emit signalAbeCamWorkerError(0);
        return;
    }
    else
    {
        m_deviceNumber = deviceNumber;
        if(m_resWidth != width || m_resHeight != height) // Mais on a changé de résolution (ou pas)
        {
            m_resWidth = width;
            m_resHeight = height;
        }
    }
    m_isCameraConnected = m_camera.open(m_deviceNumber);

    // @TODO Traiter le cas m_frame.empty()
    m_camera >> m_frame; // Pour tester
    if(m_localDebug) qDebug()<<"empty ?"<<m_frame.empty()<<"La camera est connectée ?"<<m_isCameraConnected;

    if(m_isCameraConnected)
    {
        m_inited = true;
        QTimer::singleShot(500,this, SLOT(initCapture())); // on attend un peu que la caméra soit lancée
    }
    else
    {
        // Le device n'a pas été ouvert
        emit signalAbeCamWorkerError(1); // device not open
        m_inited = false;
    }
}

void AbulEduCameraWorker::slotAbeCamWorkerSetResolution(int resWidth, int resHeight)
{
    if(m_localDebug) qDebug()  << "       *****"<<__PRETTY_FUNCTION__<<resWidth<<resHeight;
//    Q_ASSERT_X(m_inited == true, __FUNCTION__ , "Utiliser slotAbeCamWorkerSetDevice(int deviceNumber) avant d'essayer de capturer une image !");
    m_camera.set(CV_CAP_PROP_FRAME_WIDTH,  resWidth);
    m_camera.set(CV_CAP_PROP_FRAME_HEIGHT, resHeight);
    emit signalAbeCamWorkerImageSize(m_camera.get(CV_CAP_PROP_FRAME_WIDTH),m_camera.get(CV_CAP_PROP_FRAME_HEIGHT));
    /* (ré)initialise le ROI */
    slotAbeCamWorkerSetROI(0,0, m_camera.get(CV_CAP_PROP_FRAME_WIDTH), m_camera.get(CV_CAP_PROP_FRAME_HEIGHT));
    // Le device a été ouvert
    emit signalAbeCamWorkerError(0); // device open
}

void AbulEduCameraWorker::slotAbeCamWorkerGetFullImage()
{
//    if(m_localDebug) qDebug()  << "       *****"<<__FUNCTION__;
//    Q_ASSERT_X(m_inited == true, __FUNCTION__ , "Utiliser slotAbeCamWorkerSetDevice(int deviceNumber) avant d'essayer de capturer une image !");
    /* Get full Image */
//    m_time.start();
    getFrame();
    if(m_frame.data)
    {
        /* On convertit l'image de BGR en RGB
         * Plus rapide que image.rgbSwapped de Qt
         * sur ma machine, je gagne 10ms par image
         * ce qui permet du FullHD à 30FPS
         */
        cv::cvtColor(m_frame, m_frame, CV_BGR2RGB);
    }
    QImage image(m_frame.data, m_frame.cols, m_frame.rows, m_frame.step, QImage::Format_RGB888);
    emit signalAbeCamWorkerQImageReady(image);
    //    qDebug()<<m_time.elapsed();
}

void AbulEduCameraWorker::slotAbeCamWorkerGetCroppedImage()
{

}

void AbulEduCameraWorker::slotAbeCamWorkerGetFullCvMatImage()
{
//    if(m_localDebug) qDebug()  << "       *****"<<__FUNCTION__;
//    Q_ASSERT_X(m_inited == true, __FUNCTION__ , "Utiliser slotAbeCamWorkerSetDevice(int deviceNumber) avant d'essayer de capturer une image !");
    getFrame();
    emit signalAbeCamWorkerImageCvMatReady(m_frame);
}

void AbulEduCameraWorker::slotAbeCamWorkerGetCroppedCvMatImage()
{

}

void AbulEduCameraWorker::slotAbeCamWorkerSetROI(int posX, int posY, int width, int height)
{
    m_rectROI.x = posX;
    m_rectROI.y = posY;
    m_rectROI.width = width;
    m_rectROI.height = height;
}

void AbulEduCameraWorker::initCapture()
{
    if(m_localDebug) qDebug()  << "       *****"<<__FUNCTION__;
    m_image = new QImage();
    slotAbeCamWorkerSetResolution(m_resWidth, m_resHeight);
}

void AbulEduCameraWorker::getFrame()
{
//    /* L'image sera dans la matrice m_frame */

    m_camera.read(m_frame);
}
