/* AbulEduQRCodeReaderV1
 *
 * @warning aucun traitement d'erreur n'est pour l'instant implémenté
 * @see https://redmine.ryxeo.com/projects/leterrier-developpeu/repository
 * @author 2015 Jean-Louis Frucot <frucot.jeanlouis@free.fr>
 * @see The GNU Public License (GPL)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "abuleduqrcodereaderv1.h"

AbulEduQrCodeReaderV1::AbulEduQrCodeReaderV1(QObject *parent) :
    QObject(parent),
    m_scanner(0),
    m_scannerInited(false)

{
    // Permet de passer des codeBar dans les signaux
    qRegisterMetaType<codeBar>("codeBar");
    /* Normalement,on ne fait pas d'initialisations ici car cette classe est destinée à tourner dans un thread séparé
     * La communication avec la classe se fait uniquement par le biais de signaux et de slots qui sont thread-safe
     * Le point d'entrée est slotAbeCamWorkerSetDevice(int devicenumber)
     */
}

AbulEduQrCodeReaderV1::~AbulEduQrCodeReaderV1()
{
    if(m_localDebug) qDebug() << "       *****"<<__FUNCTION__<< "---------- Destructeur ----------";
    if(m_scanner)
    {
        delete m_scanner;
    }
}

void AbulEduQrCodeReaderV1::abeQrCodeReaderInit()
{
    if(!m_scannerInited)
    {
        m_scanner = new zbar::ImageScanner();
        m_scanner->set_config(zbar::ZBAR_NONE, zbar::ZBAR_CFG_ENABLE, 1);
        m_scannerInited = true;
    }
}

void AbulEduQrCodeReaderV1::slotAbeQrCodeReaderSetCvMatImage(cv::Mat frame, bool noframe)
{
    Q_ASSERT_X(m_scannerInited == true, __FUNCTION__ , "Utiliser abeQrCodeReaderInit() avant de passer une image !");
    cv::Mat grey;
    cv::cvtColor(frame,grey,CV_BGR2GRAY);
    int width = frame.cols;
    int height = frame.rows;
    uchar *raw = (uchar *)grey.data;
    /* On transforme la cv::Mat en zbar::Image */
    zbar::Image image(width, height, "Y800", raw, width * height);
    /* Scanne l'image à la recherche des QR-Codes et Codes-Barres */
    m_scanner->scan(image);
    /* On extrait les résultats qui sont pointés par symbol */
    codeBar decoded;
    for(zbar::Image::SymbolIterator symbol = image.symbol_begin();
        symbol != image.symbol_end();
        ++symbol)
    {
        std::vector<cv::Point> vp;

        decoded.type = symbol->get_type_name().c_str(); // QR-Code, EAN-xxx, ...
        decoded.data = symbol->get_data().c_str(); // Les data du code

        int n = symbol->get_location_size();
        for(int i=0;i<n;i++)
        {
            vp.push_back(cv::Point(symbol->get_location_x(i),symbol->get_location_y(i)));
        }
        cv::RotatedRect r = minAreaRect(vp);
        cv::Point2f pts[4];
        r.points(pts);
        QVector<QPair<int, int> > points;
        for(int i=0;i<4;i++)
        {
            /* On garde les points du rectangle du code barre ou QR-code dans un QVector */
            points.append(QPair<int, int>(pts[i].x, pts[i].y));
        }
        /* On garde le tout dans la structure codeBar */
        decoded.points = points;
        emit signalAbeQrCodeReaderNewDecodedDataList(decoded);
        // On dessine le rectangle qui entoure la zone décodée sauf si on n'en a pas besoin
        if(!noframe)
        {
            for(int i=0;i<4;i++)
            {
                line(frame,pts[i],pts[(i+1)%4],cv::Scalar(255,0,0),3);
            }
        }
    }
    /* Si on veut émettre ce signal car on a besoin de l'image.
     *  /!\ Le signal est émis même s'il n'y a pas de données decodées
     */
    if(!noframe)
    {
        emit signalAbeQrCodeReaderDecodedCvMatImageReady(frame);
    }
}
